/*
 *	SQLITE_MEX
 *
 *	This code is tested with SQLITE 3.11.0, MATLAB 2015b and Visual Studio
 *	Community 2015. You may need to install Visual C++ Redistributable
 *	Packages for Visual Studio 2015 from the Microsoft website. This program is
 *	distributed in the hope that it will be useful, but without any warranty.
 *
 *	Jun 17, 2015	Written by Jaewon Hwang (jaewon.hwang@gmail.com)
 *	Feb 06, 2016	Updated for SQLITE 3.11.0
 */

#include "mex.h"
#include "sqlite3.h"
#include "class_handle.h"

using namespace std;

class SQLiteDB
{
public:
	SQLiteDB();
	virtual ~SQLiteDB();

	bool Connect(const string dbfile);
	bool Disconnect();
	void Query(const string str, const mxArray* in, mxArray*& out);
	int  ErrChk(int rc);

protected:
	sqlite3* db_;
};

void GetString(const mxArray *src, string &str)
{
	size_t bufsize = mxGetNumberOfElements(src) + 1;
	char* buf = new char[bufsize];
	mxGetString(src, buf, static_cast<int>(bufsize));
	str.assign(buf);
	delete[] buf;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	size_t task = static_cast<size_t>(mxGetScalar(prhs[0]));

	switch (task)
	{
	case 0:  // version
	{
		const char *fnames[3] = { "SQLITE_VERSION", "SQLITE_VERSION_NUMBER", "SQLITE_SOURCE_ID" };
		plhs[0] = mxCreateStructMatrix(1, 1, 3, fnames);
		mxSetField(plhs[0], 0, fnames[0], mxCreateString(SQLITE_VERSION));
		mxSetField(plhs[0], 0, fnames[1], mxCreateDoubleScalar(SQLITE_VERSION_NUMBER));
		mxSetField(plhs[0], 0, fnames[2], mxCreateString(SQLITE_SOURCE_ID));
		break;
	}

	case 1:  // constructor
	{
		SQLiteDB* db = new SQLiteDB;
		if (!db) mexErrMsgTxt("SQLiteDB: Failed to create a new DB object.");
		plhs[0] = convertPtr2Mat<SQLiteDB>(db);
		break;
	}

	case 2:  // destructor
		destroyObject<SQLiteDB>(prhs[1]);
		break;

	case 3:  // connect
	{
		if (nrhs < 3 || !mxIsChar(prhs[2])) mexErrMsgTxt("SQLiteDB: DB name was not provided.");

		string dbname;  GetString(prhs[2], dbname);
		SQLiteDB *db = convertMat2Ptr<SQLiteDB>(prhs[1]);
		if (!db->Connect(dbname)) mexErrMsgTxt("SQLiteDB: Failed to connect to DB.");
		break;
	}

	case 4: // disconnect
	{
		SQLiteDB *db = convertMat2Ptr<SQLiteDB>(prhs[1]);
		if (!db->Disconnect()) mexErrMsgTxt("SQLiteDB: Failed to disconnect.");
		break;
	}

	case 5:  // query
	{
		if (nrhs < 4) mexErrMsgTxt("SQLiteDB: Missing parameter(s).");
		if (!mxIsChar(prhs[2])) mexErrMsgTxt("SQLiteDB: SQL statement was not provided.");
		string str;  GetString(prhs[2], str);
		
		SQLiteDB *db = convertMat2Ptr<SQLiteDB>(prhs[1]);
		mxArray* out = nullptr;

		try
		{
			db->Query(str, prhs[3], out);
		}
		catch (const exception& e)
		{
			mexErrMsgTxt(e.what());
		}

		plhs[0] = (out) ? out : mxCreateDoubleMatrix(0, 0, mxREAL);
		break;
	}
	}
}



SQLiteDB::SQLiteDB() : db_(nullptr)
{

}

SQLiteDB::~SQLiteDB()
{
	Disconnect();
}

bool SQLiteDB::Connect(const string dbfile)
{
	return (db_) ? true : SQLITE_OK == sqlite3_open(dbfile.c_str(), &db_);
}

bool SQLiteDB::Disconnect()
{
	bool ok = SQLITE_OK == ((db_) ? sqlite3_close(db_) : SQLITE_OK);
	if (ok) db_ = nullptr;

	return ok;
}

void SQLiteDB::Query(string str, const mxArray* in, mxArray*& out)
{
	sqlite3_stmt* sql;

	try
	{
		size_t nargin = mxGetNumberOfElements(in);
		mwIndex idx = 0;

		while (0 < str.size())
		{
			const char* tail;
			ErrChk(sqlite3_prepare_v2(db_, str.c_str(), static_cast<int>(str.size()), &sql, &tail));

			int nparam = sqlite3_bind_parameter_count(sql);
			for (int m = 1; m <= nparam; m++)			// bind index is 1-based
			{
				if (idx == nargin) throw runtime_error("SQLiteDB: insufficient bind parameter(s).");

				int rc;
				mxArray* b = mxGetCell(in, idx++);		// cell index is 0-based

				if (mxIsEmpty(b))
				{
					rc = sqlite3_bind_null(sql, m);
				}
				else
				{
					switch (mxGetClassID(b))
					{
					case mxCELL_CLASS:
						b = mxGetCell(b, 0);
						rc = sqlite3_bind_blob(sql, m, mxGetData(b), static_cast<int>(mxGetElementSize(b) * mxGetNumberOfElements(b)), SQLITE_STATIC);
						break;
					case mxCHAR_CLASS:
					{
						string p;  GetString(b, p);
						rc = sqlite3_bind_text(sql, m, p.c_str(), static_cast<int>(p.size()), SQLITE_STATIC);
						break;
					}
					case mxDOUBLE_CLASS:
					case mxSINGLE_CLASS:
						rc = sqlite3_bind_double(sql, m, mxGetScalar(b));
						break;
					case mxLOGICAL_CLASS:
					case mxINT8_CLASS:
					case mxUINT8_CLASS:
					case mxINT16_CLASS:
					case mxUINT16_CLASS:
					case mxINT32_CLASS:
					case mxUINT32_CLASS:
						rc = sqlite3_bind_int(sql, m, static_cast<int>(mxGetScalar(b)));
						break;
					case mxINT64_CLASS:
						rc = sqlite3_bind_int64(sql, m, *(static_cast<long long*>(mxGetData(b))));
						break;
					case mxUINT64_CLASS:
						rc = sqlite3_bind_int64(sql, m, static_cast<long long>(*(static_cast<unsigned long long*>(mxGetData(b)))));
						break;
					default:
						throw runtime_error("SQLiteDB: Unknown bind parameter type!!!");
						break;
					}
				}
				ErrChk(rc);
			}

			if (SQLITE_ROW == ErrChk(sqlite3_step(sql)))
			{
				// count records
				int ncol = sqlite3_column_count(sql);
				int nrow = 1;
				while (SQLITE_ROW == ErrChk(sqlite3_step(sql))) nrow++;
				ErrChk(sqlite3_reset(sql));	// rewind

				// create output matrix
				out = mxCreateCellMatrix(nrow, ncol);
				if (NULL == out) throw runtime_error("SQLiteDB: Not enough memory to create output mat.");

				// fill in the output
				for (int m = 0; m < nrow; m++)
				{
					ErrChk(sqlite3_step(sql));
					for (int n = 0; n < ncol; n++)
					{
						mxArray* a;
						switch (sqlite3_column_type(sql, n))
						{
						case SQLITE_INTEGER:
							a = mxCreateDoubleScalar(static_cast<double>(sqlite3_column_int(sql, n)));
							break;
						case SQLITE_FLOAT:
							a = mxCreateDoubleScalar(sqlite3_column_double(sql, n));
							break;
						case SQLITE_TEXT:
							a = mxCreateString(reinterpret_cast<const char*>(sqlite3_column_text(sql, n)));
							break;
						case SQLITE_BLOB:
						{
							int blobsize = sqlite3_column_bytes(sql, n);
							a = mxCreateUninitNumericMatrix(1, blobsize, mxUINT8_CLASS, mxREAL);
							memcpy(mxGetData(a), sqlite3_column_blob(sql, n), blobsize);
							break;
						}
						case SQLITE_NULL:
						default:
							a = mxCreateDoubleMatrix(0, 0, mxREAL);
							break;
						}
						mxSetCell(out, n * nrow + m, a);
					}
				}
			}

			sqlite3_finalize(sql);
			str.assign(tail);
		}
	}
	catch (...)
	{
		sqlite3_finalize(sql);
		if (out) { mxDestroyArray(out); out = nullptr; }
		rethrow_exception(current_exception());
	}
}

int SQLiteDB::ErrChk(int rc)
{
	switch (rc)
	{
	case SQLITE_OK:
	case SQLITE_ROW:
	case SQLITE_DONE:
		break;
	default:
		throw runtime_error(sqlite3_errstr(rc));
		break;
	}
	
	return rc;
}
